/* Copyright (c) 2009 The Regents of the Signet Group.
*  All rights reserved.
*
*  Permission to use, copy, modify, and distribute this software and its
*  documentation for any purpose, without fee, and without written
*  agreement is hereby granted, provided that the above copyright
*  notice, the (updated) modification history and the author appear in
*  all copies of this source code.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS `AS IS'
*  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
*  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
*  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS
*  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
*  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, LOSS OF USE, DATA,
*  OR PROFITS) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
*  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
*  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
*  THE POSSIBILITY OF SUCH DAMAGE.
*/

package net.tinyos.signet;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.security.InvalidParameterException;
import java.util.ArrayList;
import net.tinyos.message.*;
import net.tinyos.packet.*;
import net.tinyos.util.*;

/**
 * Communicates via serial port with application FlashManagerC,
 * and permits to:
 * <ul>
 * <li>Format flash
 * <li>Retrieve partition table
 * <li>Add a partition table entry
 * <li>Edit a partition's flags
 * <li>Write data in flash
 * <li>Read data from flash
 * <li>Write a file from local disk to a new partition in node flash
 * <li>Read data from node flash and store it in a file in local disk
 * <li>Read data from a partition given its id and store it in a file in local disk
 * <li>Receive progress notifications during read and write operations
 * </ul> 
 * @author Moreno Dissegna
 */
public class FlashManagerClient implements MessageListener {

	private MoteIF moteIF;	  
	private Object sem=new Object();
	private  boolean verbose;
	
	private ArrayList<PartTableEntry> part_table=new ArrayList<PartTableEntry>();
		
	private short data_buffer[];
	private int data_size;
	private int data_buffer_ptr;	
	private ReadProgressListener data_progress;

	public class PartTableEntry{
		int flags;
		int hashName;
		long size;
		long startAddr;
		long progStartOffs;	
		long spaceLeft;
		
		@Override
		public String toString() {
			return 
			"flags:         0x" + Long.toHexString(flags) + "\n"+
			"hashName:      0x" + Long.toHexString(hashName) + "\n"+
			"size:          0x" + Long.toHexString(size) + "\n"+
			"startAddr:     0x" + Long.toHexString(startAddr)+"\n"+
			"progStartOffs: 0x" + Long.toHexString(progStartOffs)+"\n"+			
			"  Space remaining: "+spaceLeft+" Bytes."
			;
		}
	}

	public interface WriteProgressListener{
		void progress(int aWrittenBytes,int aTotalBytes);
	}
	public interface ReadProgressListener{
		void progress(int aReadBytes,int aTotalBytes);
	}	
	

	public FlashManagerClient(MoteIF moteIF,boolean verbose) {
		this.verbose=verbose;
		this.moteIF = moteIF;
		this.moteIF.registerListener(new FlashManagerMessageFormat(), this);
		this.moteIF.registerListener(new FlashManagerMessageWrite(), this);
		this.moteIF.registerListener(new FlashManagerMessageTable(), this);
		this.moteIF.registerListener(new FlashManagerMessageRead(), this);
	}

	public void shutdown(){
		this.moteIF.deregisterListener(new FlashManagerMessageFormat(), this);
		this.moteIF.deregisterListener(new FlashManagerMessageWrite(), this);
		this.moteIF.deregisterListener(new FlashManagerMessageTable(), this);
		this.moteIF.deregisterListener(new FlashManagerMessageRead(), this);
	}

	private void printVerbose(String s){
		if (verbose)
			System.out.println(s);
	}

	
	
	/*
	 * BASIC COMMANDS: Send a specific message over the serial port
	 */
	
	private void sendMessage(net.tinyos.message.Message msg){
		try {			
			printVerbose("Sending command "+msg.toString());		
			moteIF.send(0, msg);	
		}catch (IOException exception) {
			System.err.println("Exception thrown when sending packets. Exiting.");
			System.err.println(exception);
		}	
	}

	private void sendFormat(){
		FlashManagerMessageFormat msg = new FlashManagerMessageFormat();
		sendMessage(msg);		  
	}

	private void sendWrite(long address,short[] data){
		FlashManagerMessageWrite msg = new FlashManagerMessageWrite();
		msg.set_address(address);
		msg.set_data(data);
		msg.set_size(data.length);		  
		sendMessage(msg);			  
	}
	private void sendSync(){
		FlashManagerMessageWrite msg = new FlashManagerMessageWrite();
		msg.set_size(0);		  
		sendMessage(msg);			  
	}
	private void sendGetTable(){
		FlashManagerMessageTable msg = new FlashManagerMessageTable();
		msg.set_cmd(FlashManagerMessagesConstants.CMD_GET_TABLE);
		sendMessage(msg);			  
	}

	private void sendAddEntry(int flags,int hashName,long size,long startAddr,int progStartOffs){
		FlashManagerMessageTable msg = new FlashManagerMessageTable();
		msg.set_cmd(FlashManagerMessagesConstants.CMD_ADD_ENTRY);
		msg.set_flags(flags);
		msg.set_hashName(hashName);
		msg.set_size(size);
		msg.set_startAddr(startAddr);
		msg.set_progStartOffs(progStartOffs);
		sendMessage(msg);			  
	}	  
	private void sendReadData(long addr,long size){
		FlashManagerMessageRead msg = new FlashManagerMessageRead();
		msg.set_address(addr);
		msg.set_size(size);
		sendMessage(msg);			  
	}
	
	
	/*
	 * Process messages received from node
	 */
	public void messageReceived(int to, Message message) {
		printVerbose("Received packet "+message.toString());
		if (message instanceof FlashManagerMessageFormat){
			synchronized(sem){
				sem.notify();
			}
		}else if (message instanceof FlashManagerMessageWrite){
			synchronized(sem){
				sem.notify();
			}
		}else if (message instanceof FlashManagerMessageTable){
			FlashManagerMessageTable fmt =(FlashManagerMessageTable)message;
			if (fmt.get_cmd()==FlashManagerMessagesConstants.CMD_GET_TABLE){
				PartTableEntry pte=new PartTableEntry();
				pte.flags		=fmt.get_flags();
				pte.hashName	=fmt.get_hashName();
				pte.size		=fmt.get_size();
				pte.startAddr	=fmt.get_startAddr();
				pte.progStartOffs =fmt.get_progStartOffs();
				pte.spaceLeft   =fmt.get_spaceLeft();
				part_table.add(pte);
			}else if (fmt.get_cmd()==FlashManagerMessagesConstants.CMD_ADD_ENTRY){
				synchronized(sem){
					sem.notify();
				}
			}
		}else if (message instanceof FlashManagerMessageRead){
			FlashManagerMessageRead msgr =((FlashManagerMessageRead) message);
			if (msgr.get_size()!=0){
				for(int i=0;i<msgr.get_size();i++){
					data_buffer[data_buffer_ptr++]=msgr.get_data()[i];	
				}
				if(data_progress!=null){
					data_progress.progress(data_buffer_ptr,data_size);
				}
			}else{
				synchronized(sem){
					sem.notify();
				}
			}
		}else{

		}
	}


	/*
	 * ADVANCED COMMANDS: 
	 * 		perform an operation sending all requested messages
	 * 		and waiting for results
	 */
	/**
	 * Formats the volume of the flash 
	 */
	public void format(){		
		try {			
			synchronized(sem){
				sendFormat();
				sem.wait(30000);
			}
			System.out.println("Format done.\n");			  
		} catch (InterruptedException e) {
			System.err.println("Timeout during operation");
			e.printStackTrace();
		}
	}
	/**
	 * @return A list containing assigned entries in the partition table
	 */
	public ArrayList<PartTableEntry> getPartitionTable(){
		part_table.clear();		
		try {
			synchronized(sem){
				sendGetTable();
				sem.wait(1000);
				// Wait 1 second, i expect to receive all the partition table in this time 
			}		  
			System.out.println("Table received.\n");
		} catch (InterruptedException e) {
		}		
		return part_table;
	}
	/**
	 * Finds a partition with given ID
	 * @param partition_table
	 * @param partitionId
	 * @return A PartitionEntry object if partition exists, null otherwise 
	 */
	public PartTableEntry getPartition(ArrayList<PartTableEntry> partition_table,int partitionId){
		for (PartTableEntry pte : partition_table) {
			if(pte.hashName==partitionId)
				return pte;
		}	
		return null;
	}
	public PartTableEntry getPartition(int partitionId){
		return getPartition(getPartitionTable(),partitionId);
	}
	/**
	 * Writes an entry at the end of the partition table
	 * @param flags
	 * @param hashName
	 * @param size
	 * @param startAddr
	 */
	public void addEntry(int flags,int hashName,long size,long startAddr,int progStartOffs){		
		try {
			synchronized(sem){
				sendAddEntry(flags, hashName, size, startAddr,progStartOffs);
				sem.wait(10000);
			}
			System.out.println("Partition table entry added.\n");			  
		} catch (InterruptedException e) {
			System.err.println("Timeout during operation");
			e.printStackTrace();
		}
	}  
	/**
	 * Edits flags of existing partition
	 * @param partitionId
	 * @param flags
	 * @throws InvalidParameterException If partition with given ID does not exist
	 */
	public void editEntryFlags(int partitionId,int flags)throws InvalidParameterException{
		if (getPartition(partitionId)==null)
			throw new InvalidParameterException();
		try {
			synchronized(sem){
				sendAddEntry(flags, partitionId, 0xFFFFFFFF, 0xFFFFFFFF,0xFFFF);
				sem.wait(10000);
			}
			System.out.println("Partition table entry edited.\n");			  
		} catch (InterruptedException e) {
			System.err.println("Timeout during operation");
			e.printStackTrace();
		}		
	}
	/** 
	 * @return Maximum number of bytes writable with writeDataChunk
	 */
	public int getWriteDataChunkMaxSize(){
		return FlashManagerMessageWrite.numElements_data();
	}
	/**
	 * Writes a chunk of data in the flash
	 * @param address  Address where data has to be written
	 * @param data     Data to be written
	 * @throws InterruptedException 
	 * @throws ArrayIndexOutOfBoundsException If length of data exceeds getWriteDataChunkMaxSize
	 */
	public void writeDataChunk(long address,short[] data) throws InterruptedException{
		if(data.length>getWriteDataChunkMaxSize())
			throw new ArrayIndexOutOfBoundsException(); 
		
		synchronized(sem){
			sendWrite(address, data);
			sem.wait(1000);
		}			  
	}
	/**
	 * Writes unlimited amount of data in the flash and ensures it's written permanently
	 * @param address
	 * @param data
	 * @param writeProgressListener Listens to progress of operation
	 */
	public void writeData(long address,byte[] data,WriteProgressListener writeProgressListener){	
		// Write data, one chunk at a time
		int chs,mchs=getWriteDataChunkMaxSize();
		short[]datach;
		try {		  
			for(int i=0;i<data.length;i+=mchs){
				chs=Math.min(data.length-i, mchs);
				datach=new short[chs];
				for(int j=0;j<chs;j++)
					datach[j]=data[i+j];				
				writeDataChunk(address, datach);	
				if (writeProgressListener!=null)
					writeProgressListener.progress(i+chs, data.length);				
				address+=chs;
			}
			syncData();
		} catch (InterruptedException e) {
			System.err.println("Timeout during image writing.");
			e.printStackTrace();
		}		  
	}	  
	public void writeData(long address,byte[] data){
		writeData(address, data,null);		
	}
	/**
	 * Ensures written data is made permanent
	 * @throws InterruptedException
	 */
	public void syncData() throws InterruptedException{		
		synchronized(sem){
			sendSync();
			sem.wait(1000);
		}			  
	}		
	
	/**
	 * Reads data from flash
	 * @param address Start read from this address
	 * @param size    Read this number of bytes
	 * @return        Data read from flash
	 */
	public short[] readData(long address,int size,ReadProgressListener aReadProgressListener){
		data_size=size;
		data_buffer=new short[size];
		data_buffer_ptr=0;
		data_progress=aReadProgressListener;
		
		try {
			synchronized(sem){
				sendReadData(address, size);
				sem.wait(120000);
				// Wait 120 seconds, i expect to receive all the requested data in this time 
			}		  
			System.out.println("Data received.\n");
		} catch (InterruptedException e) {
			System.err.println("Timeout occurred.\n");
		}		
		return data_buffer;
	}	
	public short[] readData(long address,int size){
		return readData(address,size,(ReadProgressListener)null);
	}
	/**
	 * Like previous one, but stores read data in a file.
	 * @param address
	 * @param size
	 * @param filename
	 * @param aReadProgressListener
	 * @throws IOException
	 */
	public void readData(long address,int size,String filename,ReadProgressListener aReadProgressListener) throws IOException{
		short[] data=readData(address, size, aReadProgressListener);
		FileOutputStream fos=new FileOutputStream(filename);
		byte[] out_data=new byte[size];
		for(int i=0;i<size;i++)
			out_data[i]=(byte)data[i];
		fos.write(out_data);		
	}
	public void readData(long address,int size,String filename) throws IOException{
		readData(address,size,filename,null);
	}
	
	/**
	 * Reads data of a given partition
	 * @param partitionId
	 * @param aReadProgressListener
	 * @return
	 * @throws InvalidParameterException If partition not found
	 */
	public short[] readPartition(int partitionId,ReadProgressListener aReadProgressListener){
		PartTableEntry pte=getPartition(partitionId);
		if(pte==null)
			throw new InvalidParameterException();
		return readData(pte.startAddr, (int)pte.size);
	}	
	public short[] readPartition(int partitionId){
		return readPartition(partitionId,(ReadProgressListener)null);
	}	
	/**
	 * Reads data of a given partition and stores output in file
	 * @param partitionId
	 * @param filename
	 * @param aReadProgressListener
	 * @throws IOException
	 */
	public void readPartition(int partitionId,String filename,ReadProgressListener aReadProgressListener) throws IOException{
		short[] data=readPartition(partitionId, aReadProgressListener);
		FileOutputStream fos=new FileOutputStream(filename);
		byte[] out_data=new byte[data.length];
		for(int i=0;i<data.length;i++)
			out_data[i]=(byte)data[i];
		fos.write(out_data);				
	}
	public void readPartition(int partitionId,String filename) throws IOException{
		readPartition(partitionId,filename,null);
	}	
	
	
	/**
	 * Creates a new partition entry and writes data in the newly created partition
	 * @param partitionId ID of the new partition
	 * @param data        Data to be written in the partition
	 * @param writeProgressListener Listens to progress of operation
	 * @throws InvalidParameterException If partition ID is duplicated
	 */
	public void writeNewPartition(int partitionId,byte[] data,int progStartOffs,WriteProgressListener writeProgressListener)throws InvalidParameterException{
		ArrayList<PartTableEntry> part_table=getPartitionTable();

/* ID can be duplicated, in that case the last partition is used. 
 * This allows to access to an application always by tha same id. 		
		// Check for duplicated ID
		
		if (getPartition(part_table,partitionId)!=null)
			throw new InvalidParameterException("Duplicated partition ID.");
*/			
		
		// Calculate next address 
		long addr=0;
		for (PartTableEntry pte : part_table) 			
			addr=Math.max(addr,pte.startAddr+pte.size);		

		// Insert a new partition table entry, with flags used
		int flags;
		
		flags=(
				FlashManagerPartitionMgrConstants.SPE_USED 
				) 
				^ 0xFF
				;		

		//align at page size
		if((addr % 0x100 )!=0){
			addr = addr/256;
			addr++;
			addr = addr*256;
		}		
		addEntry(flags, partitionId, data.length, addr,progStartOffs);
		getPartitionTable();
		writeData(addr,data,writeProgressListener);
		
		// After finish writing data, set working flag.
		flags=(
				FlashManagerPartitionMgrConstants.SPE_WORKING
				|
				FlashManagerPartitionMgrConstants.SPE_USED
				) 
				^ 0xFF
				;
		editEntryFlags(partitionId, flags);		
	}
	public void writeNewPartition(int partitionId,byte[] data,int progStartOffs)throws InvalidParameterException{
		writeNewPartition(partitionId, data,progStartOffs, null);		
	}
	/**
	 * Like previous method, but reads data from file.
	 * @param partitionId
	 * @param filename
	 * @param progStartOffs
	 * @param writeProgressListener
	 * @throws InvalidParameterException
	 * @throws IOException
	 */
	public void writeNewPartition(int partitionId,String filename,int progStartOffs,WriteProgressListener writeProgressListener)throws InvalidParameterException, IOException{
		
		FileInputStream fin=new FileInputStream(filename);
		
		File file=new File(filename);
		int length=(int)file.length();		
		
		fin=new FileInputStream(filename);
		byte data[]=new byte[length];
		fin.read(data);		
					
		writeNewPartition(partitionId,data,progStartOffs,writeProgressListener);		
	}
	public void writeNewPartition(int partitionId,String filename,int progStartOffs)throws InvalidParameterException, IOException{
		writeNewPartition(partitionId,filename,progStartOffs,null);
	}



	
	private static void printUsage(){
		System.err.println("usage: FlashManagerClient [-comm <source>] [-verbose] [-progress] [-print-table] [-format] [-read <read_addr> <read_size> <output filename>] [-readid <partition ID,4 digits radix 16> <output file>] [-writefile <desired partition ID, 4 digits radix 16> <local filename> <program start offset radix, 4 digits radix 16>]");
	}
	/**
	 * Command line interface to the class methods.
	 * @param args
	 * @throws Exception
	 */	
	public static void main(String[] args){	
		if(args.length==0){
			printUsage();
			System.exit(1);			
		}
		
		String source = null;
		boolean verbose=false;
		boolean format=false;
		boolean print_table=false;
		boolean read=false;
		long read_addr=0;
		int read_size=0;
		boolean write=false;
		boolean progress=false;		
		int appID=0;
		int progStartOffs=0;
		String filename="";
		String read_out_filename="";
		boolean readid=false;
		int partition_id=0;
		
		String arg;
		int argi=0;
		while(argi<args.length){
			arg=args[argi];
			if      (arg.equals("-comm")){
				argi++;
				source=args[argi];
			}else if(arg.equals("-verbose")){
				verbose=true;
			}else if(arg.equals("-progress")){
				progress=true;				
			}else if(arg.equals("-format")){
				format=true;
			}else if(arg.equals("-print-table")){
				print_table=true;
			}else if(arg.equals("-read")){
				read=true;
				read_addr=Integer.parseInt(args[++argi]);
				read_size=Integer.parseInt(args[++argi]);
				read_out_filename=args[++argi];
			}else if(arg.equals("-readid")){
				readid=true;
				partition_id=Integer.parseInt(args[++argi],16);
				read_out_filename=args[++argi];				
			}else if(arg.equals("-writefile")){
				write=true;
				appID=Integer.parseInt(args[++argi],16);
				filename=args[++argi];
				progStartOffs=Integer.parseInt(args[++argi],16);
			}else{
				System.err.println("Invalid argument: "+arg+"\n");
				printUsage();
				System.exit(1);				
			}
			argi++;
		}		
			
		PhoenixSource phoenix;

		if (source == null) {
			phoenix = BuildSource.makePhoenix(PrintStreamMessenger.err);
		}
		else {
			phoenix = BuildSource.makePhoenix(source, PrintStreamMessenger.err);
		}

		MoteIF mif = new MoteIF(phoenix);
		FlashManagerClient fmc = new FlashManagerClient(mif,verbose);

		
		if(format)
			fmc.format();
		
		if (print_table){
			for (PartTableEntry pte : fmc.getPartitionTable()) {
				System.out.println(pte.toString()+"\n");
			}
		}
		
		try{
		
		if(read){
			ReadProgressListener rpl=null;
			if(progress){
				rpl=new ReadProgressListener(){
					public void progress(int aReadBytes, int aTotalBytes) {
						System.out.println(""+aReadBytes+"B/"+aTotalBytes+"B");
					};
				};
			}
			fmc.readData(read_addr, read_size, read_out_filename, rpl);

			System.out.println("Read done");
		}
		
		if(readid){
			ReadProgressListener rpl=null;
			if(progress){
				rpl=new ReadProgressListener(){
					public void progress(int aReadBytes, int aTotalBytes) {
						System.out.println(""+aReadBytes+"B/"+aTotalBytes+"B");
					};
				};
			}
			fmc.readPartition(partition_id, read_out_filename, rpl);

			System.out.println("Read id done");			
		}
		
		if (write){		
			WriteProgressListener wpl=null;
			if(progress){
				wpl=new WriteProgressListener(){
					private int lastWritten=0;
					
					public void progress(int aWrittenBytes, int aTotalBytes) {
						//System.out.println("\\033[s"+aWrittenBytes+"B/"+aTotalBytes+"B\\033[u");
						if((1.0*(aWrittenBytes-lastWritten)/aTotalBytes)>0.1){
							System.out.println(""+aWrittenBytes+"B/"+aTotalBytes+"B");
							lastWritten=aWrittenBytes;
						}else{
							System.out.print(".");
						}
					};
				};
			}
			
			fmc.writeNewPartition(appID,filename,progStartOffs,wpl);
			
			if (print_table){
				for (PartTableEntry pte : fmc.getPartitionTable()) {
					System.out.println(pte.toString()+"\n");
				}
			}			
		}


		System.out.println("Done\n");

		fmc.shutdown();

		phoenix.shutdown();

		System.exit(0);
		}
		catch(Exception e){
			e.printStackTrace();
			System.exit(1);	
		}
	}

}


