/* Copyright (c) 2009 The Regents of the Signet Group.
*  All rights reserved.
*
*  Permission to use, copy, modify, and distribute this software and its
*  documentation for any purpose, without fee, and without written
*  agreement is hereby granted, provided that the above copyright
*  notice, the (updated) modification history and the author appear in
*  all copies of this source code.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS `AS IS'
*  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
*  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
*  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS
*  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
*  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, LOSS OF USE, DATA,
*  OR PROFITS) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
*  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
*  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
*  THE POSSIBILITY OF SUCH DAMAGE.
*/

package net.tinyos.signet;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.security.InvalidParameterException;
import java.util.Arrays;

import net.tinyos.message.*;
import net.tinyos.packet.*;
import net.tinyos.util.*;


/**
 * Communicates via serial port with application KeyVolumeManager,
 * and permits to upload and download public keys.
 * Keys file format is described in class PublicKeyRing of class TTimeSA of package SecurityTaggerV0.
 * @author Moreno Dissegna
 */
public class KeyVolumeManagerClient implements MessageListener {

	private MoteIF moteIF;	  
	private Object sem=new Object();
	private  boolean verbose;
	
	private int writeRetryCount=0;
	
	private boolean reading=false;	
	private short data_buffer[];	

	public interface WriteProgressListener{
		void progress(int aWrittenBytes,int aTotalBytes);
	}
	public interface ReadProgressListener{
		void progress(int aReadBytes,int aTotalBytes);
	}	
	

	public KeyVolumeManagerClient(MoteIF moteIF,boolean verbose) {
		this.verbose=verbose;
		this.moteIF = moteIF;
		this.moteIF.registerListener(new KeyVolumeManagerReplyPacket(), this);
	}

	public void shutdown(){
		this.moteIF.deregisterListener(new KeyVolumeManagerReplyPacket(), this);
	}

	private void printVerbose(String s){
		if (verbose)
			System.out.println(s);
	}

	
	
	/*
	 * BASIC COMMANDS: Send a specific message over the serial port
	 */
	
	private void sendMessage(net.tinyos.message.Message msg){
		try {			
			printVerbose("Sending command "+msg.toString());		
			moteIF.send(0, msg);	
		}catch (IOException exception) {
			System.err.println("Exception thrown when sending packets. Exiting.");
			System.err.println(exception);
		}	
	}

	private void sendRequest(short cmd,short[] data,int len,int offset){
		KeyVolumeManagerReqPacket req=new KeyVolumeManagerReqPacket();
		req.set_cmd(cmd);
		if(data!=null)
			req.set_data(data);
		req.set_len(len);
		req.set_offset(offset);
		sendMessage(req);
	}
	
	private void sendWrite(short[] data,int len,int offset){
		sendRequest(KeyVolumeManagerConstants.KVM_CMD_WRITE, data, len, offset);  
	}
	private void sendSync(){
		sendRequest(KeyVolumeManagerConstants.KVM_CMD_SYNC, null,0,0);			  
	}
	private void sendReadData(int offset,int length){
		KeyVolumeManagerReqPacket msg = new KeyVolumeManagerReqPacket();
		msg.set_cmd(KeyVolumeManagerConstants.KVM_CMD_READ);
		msg.set_offset(offset);
		msg.set_len(length);		
		sendMessage(msg);			  
	}
	

	
	/*
	 * Process messages received from node
	 */
	public void messageReceived(int to, Message message) {
		printVerbose("Received packet "+message.toString());
		if (message instanceof KeyVolumeManagerReplyPacket){
			KeyVolumeManagerReplyPacket rep=(KeyVolumeManagerReplyPacket)message;
			if (rep.get_error()!=0){
				System.err.println("Node signals an error occurred.");
			}

			if(reading){				
				for(int i=0;i<data_buffer.length;i++){
					data_buffer[i]=rep.get_data()[i];	
				}
			}

			synchronized(sem){
				sem.notify();
			}
		}else{

		}		
	}


	/*
	 * ADVANCED COMMANDS: 
	 * 		perform an operation sending all requested messages
	 * 		and waiting for results
	 */
	

	public short[] readDataChunk(int offset,int size) throws InterruptedException{
		reading=true;
		data_buffer=new short[size];
		
		if(size>getWriteDataChunkMaxSize())
			throw new InvalidParameterException("Size too big, maximum is:"+getWriteDataChunkMaxSize());
		
		synchronized(sem){
			sendReadData(offset, size);
			sem.wait(5000);
			// Wait 5 seconds, i expect to receive requested data in this time 
		}		  

		reading=false;
		return data_buffer;
	}		
	
	/**
	 * Reads data from flash
	 * @param address Start read from this address
	 * @param size    Read this number of bytes
	 * @return        Data read from flash
	 */
	public short[] readData(int offset,int size,ReadProgressListener aReadProgressListener){
		short[] result=new short[size];	
		try {
			int read=0;			
			while(read<size){
				int len=Math.min(getWriteDataChunkMaxSize(),size-read);
				short[] tmp=readDataChunk(offset+read, len);
				for(int i=0;i<tmp.length;i++)
					result[read+i]=tmp[i];
				read+=len;
				if(aReadProgressListener!=null)
					aReadProgressListener.progress(read, size);
			}
			System.out.println("Data received.\n");
		} catch (InterruptedException e) {
			System.err.println("Timeout occurred.\n");
		}
		return result;
	}	
	public short[] readData(int address,int size){
		return readData(address,size,(ReadProgressListener)null);
	}	

	
	/**
	 * Downloads keys stored in the node configuration storage.
	 * @param aReadProgressListener
	 * @return
	 */
	public short[] readKeys(ReadProgressListener aReadProgressListener){
		printVerbose("Retrieving key count and key length.");
		short[] result=readData(0,3);
		
		int keyCount=0,keyLength=0;
		
		keyCount+=(int)result[0];
		keyCount+=(int)result[1];
		keyLength=(int)result[2];
		printVerbose("Key count: "+keyCount+" , key length: "+keyLength);
		
		printVerbose("Downloading keys.");
		result=readData(0,2048,/*3+keyCount*keyLength*/aReadProgressListener);		
		
		return result;		
	}
	/**
	 * Like previous one, but stores read data in a file.
	 * @param filename
	 * @param aReadProgressListener
	 * @throws IOException
	 */
	public void readKeys(String filename,ReadProgressListener aReadProgressListener) throws IOException{
		short[] data=readKeys(aReadProgressListener);
		FileOutputStream fos=new FileOutputStream(filename);
		byte[] out_data=new byte[data.length];
		for(int i=0;i<data.length;i++)
			out_data[i]=(byte)data[i];
		fos.write(out_data);		
	}
	public void readKeys(int address,int size,String filename) throws IOException{
		readKeys(filename,null);
	}	
	
	/**
	 * Ensures written data is made permanent
	 * @throws InterruptedException
	 */
	public void syncData() throws InterruptedException{		
		synchronized(sem){
			sendSync();
			sem.wait(10000);
		}
	}
			
	public int getWriteDataChunkMaxSize(){
		return KeyVolumeManagerConstants.KVM_DATA_LEN;
	}	
	/**
	 * Writes a chunk of data in the flash
	 * @param offset   Offset where data has to be written
	 * @param data     Data to be written
	 * @throws InterruptedException 
	 * @throws ArrayIndexOutOfBoundsException If length of data exceeds getWriteDataChunkMaxSize
	 */
	public void writeDataChunk(int offset,short[] data) throws InterruptedException{
		if(data.length>getWriteDataChunkMaxSize())
			throw new ArrayIndexOutOfBoundsException(); 
		
		synchronized(sem){
			sendWrite(data,data.length,offset);
			sem.wait(1000);
		}			  
	}
	/**
	 * Writes unlimited amount of data in the flash and ensures it's written permanently
	 * @param address
	 * @param data
	 * @param writeProgressListener Listens to progress of operation
	 * @throws Exception 
	 */
	public void writeData(int address,byte[] data,WriteProgressListener writeProgressListener) throws Exception{	
		// Write data, one chunk at a time
		int originalAddress=address;
		int chs,mchs=getWriteDataChunkMaxSize();
		short[]datach;
				
		try {		  
			for(int i=0;i<data.length;i+=mchs){
				chs=Math.min(data.length-i, mchs);
				datach=new short[chs];
				for(int j=0;j<chs;j++)
					datach[j]=data[i+j];				
				writeDataChunk(address, datach);	
				if (writeProgressListener!=null)
					writeProgressListener.progress(i+chs, data.length);				
				address+=chs;
			}
			syncData();
			
			

			short[]writtendatashort=readData(originalAddress,data.length,null);
			byte[]writtendata=new byte[writtendatashort.length];
			for(int i=0;i<data.length;i++)
				writtendata[i]=(byte)writtendatashort[i];
			
			if(!Arrays.equals(writtendata,data)){
				writeRetryCount++;
				if(writeRetryCount<5){
					System.out.println("Written data differs from requested data. Retrying.\n");
					writeData(originalAddress,data,writeProgressListener);
				}else{		
					writeRetryCount=0;
					throw new Exception(
							"Written data differs from requested data:\n"+
							"Written data  :"+Arrays.toString(writtendata)+"\n"+
							"Requested data:"+Arrays.toString(data)+"\n"
							);
				}
			}else{
				writeRetryCount=0;
			}			
		} catch (InterruptedException e) {
			System.err.println("Timeout during image writing.");
			e.printStackTrace();
		}		  
	}	  
	public void writeData(int address,byte[] data) throws Exception{
		writeData(address, data,null);		
	}
	public void writeData(String filename,WriteProgressListener writeProgressListener) throws Exception{	
		FileInputStream fin=new FileInputStream(filename);
		
		int length=(int)(new File(filename).length());
		
		fin=new FileInputStream(filename);
		byte data[]=new byte[length];
		fin.read(data);		
		
		writeData(0, data, writeProgressListener);
	}	  
	public void writeData(String filename) throws Exception{
		writeData(filename, null);		
	}	

	
	private static void printUsage(){
		System.err.println("usage: java KeyVolumeManagerClient [-comm <source>] [-verbose] [-progress] [-upload <public key file>] [-download <output file>]");		
	}
	
	/**
	 * Command line interface to the class methods.
	 * java KeyVolumeManagerClient [-comm &lt;source&gt;] [-verbose] [-progress] [-upload &lt;public key file&gt;] [-download &lt;output file&gt;]
	 * @param args
	 * @throws Exception
	 */	
	public static void main(String[] args) throws Exception {
		String source = null;
		boolean verbose=false;
		boolean upload=false;
		boolean download=false;
		boolean progress=false;		
		String filename="";
		String downfilename="";
		
		if(args.length==0){
			printUsage();
			System.exit(1);			
		}
		
		String arg;
		int argi=0;
		while(argi<args.length){
			arg=args[argi];
			if      (arg.equals("-comm")){
				argi++;
				source=args[argi];
			}else if(arg.equals("-verbose")){
				verbose=true;
			}else if(arg.equals("-progress")){
				progress=true;					
			}else if(arg.equals("-upload")){
				upload=true;
				filename=args[++argi];
			}else if(arg.equals("-download")){
				download=true;
				downfilename=args[++argi];				
			}else{
				System.err.println("Invalid argument: "+arg+"\n");				
				printUsage();
				System.exit(1);				
			}
			argi++;
		}		
			
		PhoenixSource phoenix;

		if (source == null) {
			phoenix = BuildSource.makePhoenix(PrintStreamMessenger.err);
		}
		else {
			phoenix = BuildSource.makePhoenix(source, PrintStreamMessenger.err);
		}

		MoteIF mif = new MoteIF(phoenix);
		KeyVolumeManagerClient fmc = new KeyVolumeManagerClient(mif,verbose);


		if(upload){
			WriteProgressListener wpl=null;
			if(progress){
				wpl=new WriteProgressListener(){
					public void progress(int aWrittenBytes, int aTotalBytes) {
						System.out.println(""+aWrittenBytes+"B/"+aTotalBytes+"B");
					};
				};
			}
				
			fmc.writeData(filename,wpl);
			System.out.println("Upload done");
		}
		
		if(download){
			ReadProgressListener rpl=null;
			if(progress){
				rpl=new ReadProgressListener(){
					public void progress(int aReadBytes, int aTotalBytes) {
						System.out.println(""+aReadBytes+"B/"+aTotalBytes+"B");
					};
				};
			}
			fmc.readKeys(downfilename, rpl);

			System.out.println("Download done");			
		}
		
		System.out.println("Done\n");

		fmc.shutdown();

		phoenix.shutdown();

		System.exit(0);
	}

}


