/* Copyright (c) 2009 The Regents of the Signet Group.
*  All rights reserved.
*
*  Permission to use, copy, modify, and distribute this software and its
*  documentation for any purpose, without fee, and without written
*  agreement is hereby granted, provided that the above copyright
*  notice, the (updated) modification history and the author appear in
*  all copies of this source code.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS `AS IS'
*  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
*  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
*  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS
*  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
*  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, LOSS OF USE, DATA,
*  OR PROFITS) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
*  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
*  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
*  THE POSSIBILITY OF SUCH DAMAGE.
*/

package net.tinyos.signet;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Arrays;
import java.util.Random;

import javax.xml.parsers.ParserConfigurationException;

import org.bouncycastle.crypto.BlockCipher;
import org.bouncycastle.crypto.engines.AESEngine;
import org.bouncycastle.crypto.modes.OFBBlockCipher;
import org.bouncycastle.crypto.params.KeyParameter;
import org.bouncycastle.crypto.params.ParametersWithIV;
import org.bouncycastle.util.encoders.Hex;
import org.bouncycastle.util.encoders.HexEncoder;
import org.xml.sax.SAXException;


public class SecurityEncrypterV0 {
	private static final int BLOCK_SIZE=16;
	
	public static class EncryptionKey extends Key{
		
		public EncryptionKey(int index,int bits){
			super(0,bits/8);
		}		
		
	}
	
	public static class EncryptionKeyRing extends KeyRing<EncryptionKey>{		
		
		public static EncryptionKeyRing generate(int securityBits){
			EncryptionKey k=createNewKey(0);
			k.generate(securityBits/8);
			EncryptionKeyRing kr=new EncryptionKeyRing();
			kr.keyList.add(k);
			kr.keyNum=1;
			return kr;
		}

		protected static EncryptionKey createNewKey(int index) {
			return new EncryptionKey(index,0);
		}		
		
		@Override
		protected EncryptionKey createKey(int index) {
			return createNewKey(index);
		}
		
		public byte[] saveToBinary(){
			return keyList.get(0).value;
		}
		
		public EncryptionKey getKey(){
			return keyList.get(0);
		}

		
	}
	
	
	
	public static byte[] encrypt(byte[] data,byte[] key,byte[] iv){				
		byte[] outdata=new byte[1+BLOCK_SIZE+data.length];
		int outp=0;
		
		// put IV at the beginning
		outdata[outp++]=BLOCK_SIZE;
		System.arraycopy(iv, 0, outdata, outp, BLOCK_SIZE);
		outp+=BLOCK_SIZE;
				
		// Prepare for encryption		
		BlockCipher cypher;
		ParametersWithIV params=new ParametersWithIV(
			new KeyParameter(key)
			,
			iv
		);
		cypher=new OFBBlockCipher(new AESEngine(),BLOCK_SIZE*8);
		cypher.init(true, params);		
		
		// Block by block encryption
		int inp=0;
		int remaining=data.length;
		while(remaining>0){
			cypher.processBlock(data, inp, outdata, outp);
			inp+=BLOCK_SIZE;
			outp+=BLOCK_SIZE;
			remaining-=BLOCK_SIZE;
		}
		
		return outdata;
	}
	
	public static byte[] decrypt(byte[] data,byte[] key){		
		int ivlen;
		byte[] iv;
		
		int inp=0;
		// put IV at the beginning
		ivlen=data[inp++];
		iv=new byte[ivlen];
		System.arraycopy(data, inp, iv, 0, ivlen);
		inp+=ivlen;
		
		byte[] outdata=new byte[data.length-(1+ivlen)];		
				
		// Prepare for decryption		
		BlockCipher cypher;
		ParametersWithIV params=new ParametersWithIV(
			new KeyParameter(key)
			,
			iv
		);
		cypher=new OFBBlockCipher(new AESEngine(),BLOCK_SIZE*8);
		cypher.init(false, params);		
		
		// Block by block decryption
		int outp=0;
		int remaining=data.length-inp;
		while(remaining>0){
			cypher.processBlock(data, inp, outdata, outp);
			inp+=BLOCK_SIZE;
			outp+=BLOCK_SIZE;
			remaining-=BLOCK_SIZE;
		}
		
		return outdata;
	}	
	
	public static boolean verify(byte[] plaindata, byte[] encdata, byte[] key, int padSize){
		byte[] data=new byte[encdata.length+padSize];
		System.arraycopy(encdata, 0, data, 0, encdata.length);
		data=decrypt(data, key);
		data=Arrays.copyOf(data, data.length-padSize);
		return Arrays.equals(data, plaindata);
	}
	
	
	
	/**
	 * @param args
	 * @throws IOException 
	 * @throws ParserConfigurationException 
	 * @throws SAXException 
	 */
	public static void main(String[] args) throws IOException, SAXException, ParserConfigurationException {
		String fKeys,fnIn,fnOut;
		// Get filenames from parameters
		fKeys=args[0];
		fnIn=args[1];
		fnOut=args[2];
		
		SynapseKeyStorage sks=new SynapseKeyStorage(fKeys); 
		
		// Create files
		File fIn,fOut;		
		fIn=new File(fnIn);
		fOut=new File(fnOut);
		
		int originalSize=(int)fIn.length();
		int padSize=originalSize%BLOCK_SIZE;
		if(padSize!=0)
			padSize=BLOCK_SIZE-padSize;
		
		// Read data
		byte[] data=new byte[originalSize+padSize];
		FileInputStream fis;

		fis = new FileInputStream(fIn);
		fis.read(data);
		fis.close();					
		
		// Load key
		EncryptionKeyRing ekr= new EncryptionKeyRing();
		sks.loadEncryptionKeyRing(ekr);
		byte[] key=ekr.getKey().value;		
		
		// Generate new IV		
		byte[] iv=new byte[BLOCK_SIZE];
		new Random().nextBytes(iv);	
		
		System.out.println("Encrypting with following parameters:"); 
		System.out.println("Key : "+ekr.getKey().valueToString());
		System.out.println("IV  : "+Key.valueToString(iv));

		byte[] outdata=encrypt(data, key, iv);
		outdata=Arrays.copyOf(outdata,outdata.length-padSize);
		data=Arrays.copyOf(data,data.length-padSize);

		if(!verify(data,outdata,key,padSize))
			System.err.println("Verify failed.");
		else{		
			FileOutputStream fos;
	
			fos = new FileOutputStream(fOut);
			fos.write(outdata);
			fos.close();
		}
		
		System.out.println("Done.");
	}

}
