/* Copyright (c) 2009 The Regents of the Signet Group.
*  All rights reserved.
*
*  Permission to use, copy, modify, and distribute this software and its
*  documentation for any purpose, without fee, and without written
*  agreement is hereby granted, provided that the above copyright
*  notice, the (updated) modification history and the author appear in
*  all copies of this source code.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS `AS IS'
*  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
*  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
*  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS
*  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
*  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, LOSS OF USE, DATA,
*  OR PROFITS) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
*  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
*  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
*  THE POSSIBILITY OF SUCH DAMAGE.
*/

package net.tinyos.signet;

import java.io.File;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.security.InvalidKeyException;

import javax.xml.parsers.ParserConfigurationException;

import net.tinyos.signet.SecurityEncrypterV0.EncryptionKey;
import net.tinyos.signet.SecurityEncrypterV0.EncryptionKeyRing;

import org.bouncycastle.crypto.engines.AESEngine;
import org.xml.sax.SAXException;

public class SynapseKeyStorage {
	//private static HashFunction hf=new SHA1HashFunction();
	private static HashFunction hf=new DaviesMeyerHashConstruction(new AESEngine());
	
	private KeyStorage ks;
	
	private final static String KSID_TTIMESA="TTimeSA";
	private final static String KSID_CC2420ENCRYPTION="CC2420Encryption";
	private final static String KSID_CC2420_DOS_AUTHENTICATION="CC2420DosAuthentication";
	private final static String KSID_CC2420_DOS_SESSION="CC2420DosSession";
	
	public SynapseKeyStorage(String filename) throws SAXException, IOException, ParserConfigurationException{		
		ks=new KeyStorage(filename);		
	}
	
	public void generate(int authSecurity, int authUses, int encryptionSecurity, int dosAuthSecurity) throws ParserConfigurationException, IOException{
		// Generate key for TTimeSA
		TTimeSA sa=new TTimeSA(hf);
		ks.addKeyRing(
				KSID_TTIMESA, 
				sa.generateNewKeyRing(authSecurity/8, authUses)
				);
		
		// Generate key for encryption		
		ks.addKeyRing(
				KSID_CC2420ENCRYPTION, 
				SecurityEncrypterV0.EncryptionKeyRing.generate(encryptionSecurity)
				);
		
		// Generate key for dos protection authentication		
		ks.addKeyRing(
				KSID_CC2420_DOS_AUTHENTICATION, 
				SecurityEncrypterV0.EncryptionKeyRing.generate(dosAuthSecurity)
				);	
		
		// Generate DoS session counter		
		ks.addKeyRing(
				KSID_CC2420_DOS_SESSION, 
				DosSessionKeyRing.generate()
				);		
		
		// Save
		ks.save();
	}
	
	public byte[] export() throws InvalidKeyException, ParserConfigurationException{
		
		TTimeSA sa=new TTimeSA(hf);
		TTimeSA.SecretKeyRing skr=sa.generateNewKeyRing(1, 1);
		loadTTimeSASecretKeyRing(skr);		
		byte ttimesa[]=skr.getPublishableKeyRing().saveToBinary();
		
		EncryptionKeyRing ekr=new EncryptionKeyRing();
		loadEncryptionKeyRing(ekr);
		byte enc[]=ekr.saveToBinary();
		
		EncryptionKeyRing dakr=new EncryptionKeyRing();
		loadDosAuthenticationKeyRing(dakr);
		byte dosauth[]=dakr.saveToBinary();
		
		DosSessionKeyRing dskr=new DosSessionKeyRing();
		loadDosSessionKeyRing(dskr);
		byte dossess[]=dskr.saveToBinary();		
		
		byte output[]=new byte[ttimesa.length + enc.length + dosauth.length + dossess.length];
		
		int i=0;
		System.arraycopy(ttimesa, 0, output, i, ttimesa.length);
		i+=ttimesa.length;
		System.arraycopy(enc, 0, output, i, enc.length);
		i+=enc.length;
		System.arraycopy(dosauth, 0, output, i, dosauth.length);
		i+=dosauth.length;		
		System.arraycopy(dossess, 0, output, i, dossess.length);
		i+=dossess.length;
		
		return output;
	}
	
	public void export(String filename) throws InvalidKeyException, IOException, ParserConfigurationException{		
		byte output[]=export();
		FileOutputStream fos=new FileOutputStream(filename);
		fos.write(output);
		fos.close();
	}	
	
	public void loadTTimeSASecretKeyRing(TTimeSA.SecretKeyRing skr) throws ParserConfigurationException{		
		skr.loadFromXML(ks.getKeyRingContainer(KSID_TTIMESA));
	}
	
	public void loadEncryptionKeyRing(EncryptionKeyRing ekr) throws ParserConfigurationException{
		ekr.loadFromXML(ks.getKeyRingContainer(KSID_CC2420ENCRYPTION));		
	}
	
	public void loadDosAuthenticationKeyRing(EncryptionKeyRing dakr) throws ParserConfigurationException{
		dakr.loadFromXML(ks.getKeyRingContainer(KSID_CC2420_DOS_AUTHENTICATION));		
	}	
	
	public void loadDosSessionKeyRing(DosSessionKeyRing dskr) throws ParserConfigurationException{
		dskr.loadFromXML(ks.getKeyRingContainer(KSID_CC2420_DOS_SESSION));		
	}		
	
	public static class DosSessionKeyRing extends KeyRing<Key>{		
		
		public static DosSessionKeyRing generate(){
			Key k=createNewKey(0);
			DosSessionKeyRing kr=new DosSessionKeyRing();
			kr.keyList.add(k);
			kr.keyNum=1;
			return kr;
		}
		private static Key createNewKey(int index) {
			byte value[]={(byte)0,(byte)0};
			return new Key(index,value);
		}			
		@Override
		protected Key createKey(int index) {
			return createNewKey(index);
		}		
		public byte[] saveToBinary(){
			return keyList.get(0).value;
		}		
		public Key getKey(){
			return keyList.get(0);
		}		
	}	
	
	
	
	public static void printUsage(){
		System.err.println("Usage: java SynapseKeyStorage [-generate <# of security bits for authentication> <#of uses per key> <# of security bits for encryption> <# of security bits for DoS protection> <filename>] [-get-public <key storage filename> <destination filename>]");
		System.exit(1);
	}
	
	/**
	 * @param args
	 * @throws ParserConfigurationException 
	 * @throws IOException 
	 * @throws SAXException 
	 * @throws InvalidKeyException 
	 */
	public static void main(String[] args) throws SAXException, IOException, ParserConfigurationException, InvalidKeyException {
		boolean generate=false,export=false;
		int authSecurity=0,authUses=0,encSecurity=0,dosSecurity=0;		
		String newKeysFile="",keyStorageFile="",exportFile="";
		
		int argi=0;		
		while(argi<args.length){
			if      (args[argi].equals("-generate")){
				generate=true;
				authSecurity=Integer.parseInt(args[++argi]);
				authUses=Integer.parseInt(args[++argi]);
				encSecurity=Integer.parseInt(args[++argi]);
				dosSecurity=Integer.parseInt(args[++argi]);
				newKeysFile=args[++argi];
			}else if (args[argi].equals("-get-public")){
				export=true;
				keyStorageFile=args[++argi];
				exportFile=args[++argi];
			}else
				printUsage();
			argi++;
		}
		if(generate){
			SynapseKeyStorage sks=new SynapseKeyStorage(newKeysFile);
			sks.generate(authSecurity,authUses,encSecurity,dosSecurity);
		}
		if(export){
			SynapseKeyStorage sks=new SynapseKeyStorage(keyStorageFile);
			sks.export(exportFile);
		}
	}

}
