/* Copyright (c) 2009 The Regents of the Signet Group.
*  All rights reserved.
*
*  Permission to use, copy, modify, and distribute this software and its
*  documentation for any purpose, without fee, and without written
*  agreement is hereby granted, provided that the above copyright
*  notice, the (updated) modification history and the author appear in
*  all copies of this source code.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS `AS IS'
*  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
*  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
*  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS
*  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
*  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, LOSS OF USE, DATA,
*  OR PROFITS) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
*  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
*  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
*  THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef SECURITY_TAGS_V0_H
#define SECURITY_TAGS_V0_H

#include "TTimeSA.h"
#include "debug_printf.h"

/**
 * Definition of file format:
 * 
 *   +---------+-------------\+----------------+------------------\-------------\---------------------
 *   | Sig.len |  Signature  /|  SessionInfo   | Hashes of blocks /   New keys  /      Remaining data 
 *   | uint8_t | hash_v0_t[] \| ST0SessionInfo |     hash_v0_t[]  \ ST0NewKey[] \         uint8_t[]         
 *   +---------+-------------/+----------------+------------------/-------------/---------------------
 *   ^                                                      ^                 ^
 *   |______________________________________________________|_________________|_____________...
 *          ST0BootstrapBlock  - First transport block                   Other blocks
 * 
 * Signature is the signature of SessionInfo, Hashes of blocks, and Data pad.
 * The number of items in Signature is specified in the first byte of the block 
 * The number of items in Hashes of blocks is specified in SessionInfo, it's blocksCount
 * 
 * SessionInfo contains hashOnesAdjust, one byte selected to minimize number of ones in the hash 
 * of signed data, to minimize size of the signature. 
 * In this way, having 256 ways to select this byte we can have good chances to avoid having a high 
 * number of ones in the hash and thus having a very long signature. 
 */
 
#ifndef BLOCK_PAYLOAD_SIZE
    #define BLOCK_PAYLOAD_SIZE 800
#endif      
  
enum{
    ST0_TB_SIZE=BLOCK_PAYLOAD_SIZE, 
    ST0_MAX_TRANFER_SIZE=48, // Maximum size of data to be transferred, in KB
    ST0_MAX_BLOCKS=(uint16_t)((uint16_t)ST0_MAX_TRANFER_SIZE*(uint16_t)1024/(uint16_t)ST0_TB_SIZE), // Blocks needed to transfer a 43K app 
    ST0_HASH_SIZE=SIG_SEC_PARAM,  // Number of bytes for one hash  
}; 
 
/*
 * Structures contained in file
 * IMPORTANT: This structures MUST be nx_struct because they must be packed and big-endian
 */ 
 
typedef nx_struct ST0Header{
	nx_uint8_t signatureLength; // Number of HASHES, not of bytes
}ST0Header;
 
enum{
	ST0_HEADER_SIZE=sizeof(ST0Header),
};
 
typedef nx_struct ST0SessionInfo{       
    nx_uint8_t    blocksCount;       // Number of transport blocks needed
    nx_uint8_t    newKeyCount;       // Number of new keys included 
    nxle_uint32_t codeImageSize;     // Size of code image only
    nx_uint8_t    hashOnesAdjust;    // One byte selected to minimize number of ones in signed data hash        
}ST0SessionInfo;

enum{
    ST0_SESSION_INFO_SIZE=sizeof(ST0SessionInfo),                       
};

typedef nx_uint8_t hash_v0_t[ST0_HASH_SIZE];


typedef nx_struct ST0NewKey{
	nx_uint8_t keyIndex;   // Index of the key
	hash_v0_t  key;        // Value of the key
}ST0NewKey;

enum{
    ST0_NEW_KEY_SIZE=sizeof(ST0NewKey),                       
};

/* Structure containing infos and pointers to all significant parts of bootstrap block.
 * This structure don't have to be nx_struct because it's not packed in the file.
 * This structure is instead built reading file.
 */
typedef struct ST0BootstrapBlock{
	// Pointer to beginning of block
	uint8_t        *block;
	// Pointers to main parts	
	ST0Header      *header;
	hash_v0_t      *signature;
    ST0SessionInfo *sessionInfo;
    hash_v0_t      *hashesOfBlocks;    
//    uint8_t        *dataPad;
    
    // Pointer to and length of signed data
    uint8_t        *signedData;
    uint16_t       signedDataLength;
    
    // Size of hashes stored in first block
    uint16_t       hashesOfBlocksSize;
       
    // Size of security tags
    uint16_t       tagsSize; 
    
    bool            valid;
}ST0BootstrapBlock;


/**
 * Fills ST0BootstrapBlock structure with information retrieved from block.
 */
void loadST0BootstrapBlock(ST0BootstrapBlock *blockData,void *block){
	nx_uint8_t *p=block;
	
	blockData->valid=FALSE;
	
	blockData->block=(uint8_t*)p;
	
	blockData->header = (ST0Header*)(void*)p;
	p=p + ST0_HEADER_SIZE;
	
	// Check signature length
	// 1: Number of hashes < maximum
	if(blockData->header->signatureLength>SIG_BITVEC_BITSIZE){
	   return;
	} 
	// 2: minimum size of block obtained with this signature < max size of block
	if((
	   ST0_HEADER_SIZE
	   +
	   ST0_HASH_SIZE * blockData->header->signatureLength
	   +
	   ST0_SESSION_INFO_SIZE	   
	)>ST0_TB_SIZE){
	   return;
	}
	
	blockData->signature = (hash_v0_t*)(void*)p;
	p=p + ST0_HASH_SIZE * blockData->header->signatureLength;

    blockData->sessionInfo = (ST0SessionInfo*)(void*)p;
    p=p + ST0_SESSION_INFO_SIZE;	
    
    // Check session info
    // 1: Number of blocks
    if(blockData->sessionInfo->blocksCount>ST0_MAX_BLOCKS){
        return;
    }
    // 2: Code image size
    {
    // Use this because comparison between nx_uint32 and uint16 does not work
    uint16_t size=(uint16_t)blockData->sessionInfo->codeImageSize;
    if(size>((uint16_t)((uint16_t)ST0_MAX_TRANFER_SIZE*(uint16_t)1024))){
        return;
    }
    }        
    // 3: # of new keys
    if(blockData->sessionInfo->newKeyCount>SIG_BITVEC_BITSIZE){
        return;
    }
    
    blockData->hashesOfBlocks = (hash_v0_t*)(void*)p;
    p=p + ST0_HASH_SIZE * blockData->sessionInfo->blocksCount;
       
    blockData->signedData = (uint8_t*)(void*)blockData->sessionInfo;
    blockData->signedDataLength = 
        ST0_TB_SIZE 
        - 
        ST0_HASH_SIZE * blockData->header->signatureLength 
        - 
        ST0_HEADER_SIZE
    ;  
    
    blockData->hashesOfBlocksSize =
        blockData->signedDataLength
        -
        ST0_SESSION_INFO_SIZE
    ;
    if (blockData->hashesOfBlocksSize > ST0_HASH_SIZE * blockData->sessionInfo->blocksCount)
        blockData->hashesOfBlocksSize = ST0_HASH_SIZE * blockData->sessionInfo->blocksCount;
        
        
        
    blockData->tagsSize =
        ST0_HEADER_SIZE
        +
        ST0_HASH_SIZE * blockData->header->signatureLength
        +
        ST0_SESSION_INFO_SIZE
        +
        ST0_HASH_SIZE * blockData->sessionInfo->blocksCount
        +
        ST0_NEW_KEY_SIZE * blockData->sessionInfo->newKeyCount
    ;     
    
    blockData->valid=TRUE;
}



#endif /* SECURITY_TAGS_V0_H */
